"""Testing Change Data Methods.
"""
from changelist_sort.change_data import ChangeData
from test import data_provider


def test_get_sort_path_after_src_file_returns_after():
    expected = '/app/src/main/java/app/Main.java'
    test_data = ChangeData(
        after_dir=False,
        after_path=expected
    )
    result = test_data._get_sort_path()
    assert expected == result


def test_get_sort_path_before_and_after_src_files_returns_after():
    expected = '/app/src/main/java/app/Main.java'
    test_data = ChangeData(
        before_dir=False,
        before_path='/app/src/main/java/app/File.java',
        after_dir=False,
        after_path=expected
    )
    result = test_data._get_sort_path()
    assert expected == result


def test_get_sort_path_before_src_file_returns_before():
    expected = '/app/src/main/java/app/Main.java'
    test_data = ChangeData(
        before_dir=False,
        before_path=expected,
    )
    result = test_data._get_sort_path()
    assert expected == result


def test_first_dir_root_gradle_returns_none():
    cd = data_provider.get_root_gradle_build_change_data()
    assert cd.first_dir is None


def test_first_dir_app_gradle_returns_app():
    cd = data_provider.get_app_gradle_build_change_data()
    assert 'app' == cd.first_dir


def test_first_dir_module_src_returns_module():
    cd = data_provider.get_module_src_change_data()
    assert 'module' == cd.first_dir


def test_first_dir_module_test_returns_module():
    cd = data_provider.get_module_test_change_data()
    assert 'module' == cd.first_dir


def test_first_dir_gradle_properties_returns_gradle():
    cd = data_provider.get_gradle_properties_change_data()
    assert 'gradle' == cd.first_dir


def test_first_dir_github_workflows_returns_github():
    cd = data_provider.get_github_workflows_change_data()
    assert '.github' == cd.first_dir


def test_file_basename_module_src_file_returns_file():
    cd = data_provider.get_module_src_change_data()
    assert 'Main.java' == cd.file_basename


def test_file_basename_gradle_properties_returns_file():
    cd = data_provider.get_gradle_properties_change_data()
    assert 'gradle-wrapper.properties' == cd.file_basename


def test_file_basename_root_gradle_build_returns_file():
    cd = data_provider.get_root_gradle_build_change_data()
    assert 'build.gradle' == cd.file_basename


def test_file_ext_module_src_file_returns_ext():
    cd = data_provider.get_module_src_change_data()
    assert 'java' == cd.file_ext


def test_file_ext_root_gradle_build_returns_ext():
    cd = data_provider.get_root_gradle_build_change_data()
    assert 'gradle' == cd.file_ext


def test_file_ext_github_src_file_returns_ext():
    cd = data_provider.get_github_workflows_change_data()
    assert 'yml' == cd.file_ext


def test_file_ext_hidden_file_with_ext_returns_ext():
    cd = data_provider.get_change_data('/module/.hidden_file.c')
    assert 'c' == cd.file_ext


def test_file_ext_hidden_file_no_ext_returns_none():
    cd = data_provider.get_change_data('/module/.hidden_file')
    assert cd.file_ext is None


def test_file_ext_gradle_kts_returns_ext():
    cd = data_provider.get_change_data('/module/build.gradle.kts')
    assert 'gradle.kts' == cd.file_ext



""" Testing Changelist Map
"""
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.changelist_map import ChangelistMap
from test import data_provider


def test_insert_simple_returns_true():
    instance = ChangelistMap()
    assert instance.insert(ChangelistData(
        id='1234',
        name='Build Updates',
    ))


def test_insert_twice_same_id_returns_false():
    instance = ChangelistMap()
    same_id = instance._generate_new_id()
    assert instance.insert(ChangelistData(
        id=same_id,
        name='Module',
    ))
    assert not instance.insert(ChangelistData(
        id=same_id,
        name='App',
    ))


def test_insert_twice_different_id_same_name_returns_false():
    instance = ChangelistMap()
    init_cl = data_provider.get_module_changelist()
    assert instance.insert(init_cl)
    # Insert New CL with different id, same name
    new_cl = ChangelistData(
        id=instance._generate_new_id(),
        name=init_cl.name,
    )
    assert not instance.insert(new_cl)


def test_insert_gradle_changelists_returns():
    instance = ChangelistMap()
    gradle_cl = data_provider.get_multiple_gradle_changelists()
    assert len(gradle_cl) == 2
    assert instance.insert(gradle_cl[0])
    assert not instance.insert(gradle_cl[1])    # Has the Same Changelist Name
    #
    result = instance.search(gradle_cl[0].list_key.key)
    assert result.name == gradle_cl[0].name # The first in the list is returned


def test_search_empty_returns_none():
    instance = ChangelistMap()
    assert instance.search(' ') is None
    assert instance.search('') is None
    assert instance.search('a') is None
    assert instance.search('root') is None
    assert instance.search('project root') is None
    assert instance.search('gradle') is None


def test_search_simple_returns_changelist():
    instance = ChangelistMap()
    assert instance.insert(ChangelistData(
        id='1234',
        name='Build Updates',
    ))
    assert instance.search('buildupdates') is not None


def test_contains_id_empty_returns_false():
    instance = ChangelistMap()
    assert not instance.contains_id('1234')


def test_contains_id_simple_returns_true():
    instance = ChangelistMap()
    assert instance.insert(ChangelistData(
        id='1234',
        name='Build Updates',
    ))
    assert instance.contains_id('1234')


def test_get_lists_empty_returns_empty():
    instance = ChangelistMap()
    assert 0 == len(instance.get_lists())


def test_get_lists_simple_returns_list():
    instance = ChangelistMap()
    assert instance.insert(ChangelistData(
        id='1234',
        name='Build Updates',
    ))
    assert 1 == len(instance.get_lists())


def test_generate_new_id_then_insert_times_n():
    instance = ChangelistMap()
    for i in range(100):
        result = instance._generate_new_id()
        assert 36 == len(result)
        assert not instance.contains_id(result)
        assert instance.insert(
            ChangelistData(
                id=result,
                name=f'{i}',    # A new name is required each time too
            )
        )
        assert instance.contains_id(result)


def test_create_changelist_invalid_type():
    instance = ChangelistMap()
    try:
        instance.create_changelist(1)
        type_error_raised = False
    except TypeError:
        type_error_raised = True
    assert type_error_raised


def test_create_changelist_same_name_twice():
    instance = ChangelistMap()
    new_cl_name = 'Main Files'
    first_cl = instance.create_changelist(new_cl_name)
    second_cl = instance.create_changelist(new_cl_name)
    assert first_cl == second_cl
    assert first_cl.id == second_cl.id



""" Sort With Developer File Patterns.
"""
from changelist_sort import list_key
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.list_key import ListKey
from changelist_sort.sorting import file_sort, module_sort
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.workspace.developer_changelist import DeveloperChangelist
from changelist_sort.workspace.developer_file_pattern import DeveloperFilePattern


_SRC_DIR_PATTERN = DeveloperFilePattern(first_dir='changelist_sort')
_TEST_DIR_PATTERN = DeveloperFilePattern(first_dir='test')
_INPUT_PACKAGE_PATTERN = DeveloperFilePattern(path_end='input')
_SORTING_PACKAGE_PATTERN = DeveloperFilePattern(path_end='sorting')
_WORKSPACE_PACKAGE_PATTERN = DeveloperFilePattern(path_end='workspace')

_BUILD_UPDATES_KEY = list_key.compute_key('Build Updates')

# Modify these Patterns
DEVELOPER_CL_TUPLE: tuple[DeveloperChangelist, ...] = (
    DeveloperChangelist(
        None,
        list_key.compute_key('Documentation'),
        (
            DeveloperFilePattern(
                file_ext='md',
            ),
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Input Package Tests'),
        (
            _TEST_DIR_PATTERN,
            _INPUT_PACKAGE_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Sorting Package Tests'),
        (
            _TEST_DIR_PATTERN,
            _SORTING_PACKAGE_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Workspace Package Tests'),
        (
            _TEST_DIR_PATTERN,
            _WORKSPACE_PACKAGE_PATTERN,
        )
    ),
    # Tests that don't match a pattern
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Tests'),
        (
            _TEST_DIR_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Input Package'),
        (
            _SRC_DIR_PATTERN,
            _INPUT_PACKAGE_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Sorting Package'),
        (
            _SRC_DIR_PATTERN,
            _SORTING_PACKAGE_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Workspace Package'),
        (
            _SRC_DIR_PATTERN,
            _WORKSPACE_PACKAGE_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.MODULE,
        list_key.compute_key('Main Package Source'),
        (
            _SRC_DIR_PATTERN,
        )
    ),
    DeveloperChangelist(
        ModuleType.ROOT,
        list_key.compute_key('Shell Scripts'),
        (
            DeveloperFilePattern(
                file_ext='sh',
            ),
        )
    ),
    DeveloperChangelist(
        ModuleType.ROOT,
        list_key.compute_key('Project Root'),
        (
            DeveloperFilePattern(
                inverse=True,
                first_dir='gradle',
            ),
            DeveloperFilePattern(
                inverse=True,
                file_ext='gradle',
            ),
            DeveloperFilePattern(
                inverse=True,
                file_ext='kts',
            ),
        )
    ),
    DeveloperChangelist(
        ModuleType.GRADLE,
        _BUILD_UPDATES_KEY,
        (
            DeveloperFilePattern(
                inverse=True,
                first_dir='gradle',
            ),
            DeveloperFilePattern(
                inverse=True,
                first_dir=None,
            ),
        )
    ),
    DeveloperChangelist(
        ModuleType.ROOT,
        _BUILD_UPDATES_KEY,
        (
            DeveloperFilePattern(file_ext='gradle'),
        )
    ),
    DeveloperChangelist(
        ModuleType.ROOT,
        _BUILD_UPDATES_KEY,
        (
            DeveloperFilePattern(file_ext='properties'),
        )
    ),
    DeveloperChangelist(
        ModuleType.GRADLE,
        list_key.compute_key('Module Gradle Build Files'),
        (
            DeveloperFilePattern(
                inverse=True,
                first_dir='gradle',
            ),
            DeveloperFilePattern(
                inverse=True,
                first_dir=None,
            ),
        )
    ),
)


def _filter_patterns_by_module(
    module_type: ModuleType | None
) -> tuple[DeveloperChangelist, ...]:
    """
    Filter the Changelists by the ModuleType their Pattern applies to.
    """
    return tuple(filter(
        lambda dcl: dcl.module_type is None or dcl.module_type == module_type,
        DEVELOPER_CL_TUPLE
    ))


def sort_file_by_developer(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    """
    Apply the Developer FilePattern Setting to Sort a single File into the Changelist Map.
    - Filters Patterns by matching ModuleType before checking files.
    - Fallback to Module Sort
    """
    # Filter Developer Changelist Tuple by File's ModuleType 
    filtered_dcl_patterns = _filter_patterns_by_module(file_sort.get_module_type(file))
    # Check Developer Changelists in Tuple Order
    for dcl_pattern in filtered_dcl_patterns:
        if dcl_pattern.check_file(file):
            # Pattern Matched.
            # Search Map. Add File to Changelist.
            if (cl := cl_map.search(dcl_pattern.list_key.key)) is not None:
                cl.changes.append(file)
                return True
            # Create the Developer Changelist. Add File to Changelist.
            cl_map.create_changelist(dcl_pattern.list_key.changelist_name).changes.append(file)
            return True
    # Fallback to Module Sort when Developer Sort Fails.
    return module_sort.sort_file_by_module(cl_map, file)


def is_sorted_by_developer(
    changelist_key: ListKey,
    file: ChangeData,
) -> bool:
    """
    Determines if this File matches the ChangeList Key or Name.
    - Finds the First DeveloperChangelist Pattern that matches
    - Fallback to Module Sort
    """
    # Filter Developer Changelist Tuple by File's ModuleType 
    filtered_dcl_patterns = _filter_patterns_by_module(file_sort.get_module_type(file))
    # Check Developer Changelists in Tuple Order
    for dcl_pattern in filtered_dcl_patterns:
        if dcl_pattern.check_file(file):
            # Pattern Matched
            if dcl_pattern.list_key.key == changelist_key.key or\
                dcl_pattern.list_key.changelist_name == changelist_key.changelist_name:
                return True
            # This File could be sorted higher in the Developer Changelist order.
            return False
    # Fallback to Module Sort.
    return module_sort.is_sorted_by_module(changelist_key, file)



""" Developer Changelist Settings and Methods
"""
from dataclasses import dataclass

from changelist_sort.change_data import ChangeData
from changelist_sort.list_key import ListKey
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.workspace.developer_file_pattern import DeveloperFilePattern


@dataclass(frozen=True)
class DeveloperChangelist:
    """
    The Changelist Defined by the Developer.

    Parameters:
    - list_key (ListKey): The Key and Name of the Changelist.
    - module_type (ModuleType): The Type of Module to match Files with.
    - file_patterns (tuple[DeveloperFilePattern]): 
    """
    module_type: ModuleType | None
    list_key: ListKey
    file_patterns: tuple[DeveloperFilePattern]

    def check_file(self, file: ChangeData) -> bool:
        """
        Determine if the File can be added to this Changelist.

        Parameters:
        - file (ChangeData): The ChangeData of the File to pattern match.

        Returns:
        True if the File matches all patterns in this Changelist.
        """
        for fp in self.file_patterns:
            if not fp.check_file(file):
                return False
        return True



""" Main Package Methods.
"""
from changelist_sort import workspace
from changelist_sort.sorting import sort
from changelist_sort.input.input_data import InputData


def sort_changelists(
    input_data: InputData,
):
    """
    Sort the given Changelists and write them to the Workspace File.
    """
    ws_tree = workspace.get_workspace_tree(input_data)
    sorted_lists = sort(
        ws_tree.extract_list_elements(),
        input_data.sort_mode
    )
    ws_tree.replace_changelists(sorted_lists)
    workspace.write_workspace_tree(
        ws_tree, input_data.workspace_path
    )



""" Testing Module Sort Methods.
"""
import pytest

from test import data_provider
from changelist_sort.sorting import file_sort
from changelist_sort.sorting import module_sort
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.sorting.module_sort import sort_file_by_module, is_sorted_by_module


@pytest.mark.parametrize(
    'key', [ModuleType.MODULE, ModuleType.HIDDEN]
)
def test_get_module_keys_user_modules(key):
    result = module_sort.get_module_keys(key)
    assert result == tuple()


def test_sort_file_by_module_empty_map_empty_file_returns_false():
    cl_map = ChangelistMap()
    empty_file = ChangeData()
    assert not sort_file_by_module(cl_map, empty_file)


def test_sort_file_by_module_empty_map_module_src_file_returns_true():
    cl_map = ChangelistMap()
    src_file = data_provider.get_change_data(
        data_provider.MODULE_SRC_PATH,
    )
    assert sort_file_by_module(cl_map, src_file)
    # Check CL Map for new Changelist
    new_cl = cl_map.search('module')
    assert new_cl is not None
    assert new_cl.name == 'Module'


def test_sort_file_by_module_app_gradle_file_returns_true_inserts_build_updates():
    cl_map = ChangelistMap()
    test_file = data_provider.get_app_gradle_build_change_data()
    assert sort_file_by_module(cl_map, test_file)
    # Check CL Map for new Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert len(new_cl.changes) == 1
    assert new_cl.name == 'Build Updates'


def test_sort_file_by_module_gradle_properties_returns_true_inserts_build_updates():
    cl_map = ChangelistMap()
    test_file = data_provider.get_app_gradle_build_change_data()
    assert sort_file_by_module(cl_map, test_file)
    # Check CL Map for new Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert len(new_cl.changes) == 1
    assert new_cl.name == 'Build Updates'


def test_sort_file_by_module_github_workflows_returns_true_inserts_github():
    cl_map = ChangelistMap()
    test_file = data_provider.get_github_workflows_change_data()
    assert sort_file_by_module(cl_map, test_file)
    # Check CL Map for new Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert len(new_cl.changes) == 1
    assert new_cl.name == 'Github'


def test_sort_file_by_module_empty_file_returns_false():
    cl_map = ChangelistMap()
    empty_file = ChangeData()
    assert not sort_file_by_module(cl_map, empty_file)


def test_sort_file_by_module_app_cl_app_gradle_returns_true():
    cl_map = ChangelistMap()
    cl_map.insert(data_provider.get_app_changelist())
    new_cd = data_provider.get_app_gradle_build_change_data()
    assert sort_file_by_module(cl_map, new_cd)
    # A new Changelist is created called Build Updates
    result = cl_map.get_lists()
    assert len(result) == 2
    cl_0 = result[0]
    assert cl_0.name == 'App'
    cl_1 = result[1]
    assert cl_1.name == 'Build Updates'
    assert new_cd in cl_1.changes


def test_sort_file_by_module_build_updates_cl_app_gradle_returns_true():
    cl_map = ChangelistMap()
    cl_map.insert(data_provider.get_build_updates_changelist())
    new_cd = data_provider.get_app_gradle_build_change_data()
    assert sort_file_by_module(cl_map, new_cd)
    # The Build Updates Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    assert result[0].name == 'Build Updates'


def test_sort_file_by_module_root_cl_readme_returns_true():
    cl_map = ChangelistMap()
    cl_map.insert(data_provider.get_root_changelist())
    new_cd = data_provider.get_root_readme_change_data()
    assert sort_file_by_module(cl_map, new_cd)
    # The Build Updates Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    assert result[0].name == 'Root'
    assert len(result[0].changes) == 1


def test_sort_file_by_module_module_cl_module_src_returns_true():
    cl_map = ChangelistMap()
    cl_map.insert(data_provider.get_module_changelist())
    new_cd = data_provider.get_module_src_change_data()
    assert sort_file_by_module(cl_map, new_cd)
    # The Src file is added to the existing changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    cl_0 = result[0]
    assert cl_0.name == 'Module'
    assert new_cd in cl_0.changes


def test_sort_file_by_module_zero_len_module_returns_false():
    new_cd = data_provider.get_change_data('//hello.py')
    assert not sort_file_by_module(None, new_cd)


def test_sort_file_by_module_root_readme_returns_true():
    cl_map = ChangelistMap()
    new_cd = data_provider.get_root_readme_change_data()
    assert file_sort.get_module_type(new_cd) == ModuleType.ROOT
    assert sort_file_by_module(cl_map, new_cd)


def test_is_sorted_by_module_module_cl():
    cl = data_provider.get_module_changelist()
    for file in cl.changes:
        assert is_sorted_by_module(cl.list_key, file)


def test_is_sorted_by_module_app_cl_app_gradle_returns_false():
    cl = data_provider.get_app_changelist()
    assert not is_sorted_by_module(
        cl.list_key, data_provider.get_app_gradle_build_change_data()
    )


def test_is_sorted_by_module_app_cl_strings_res_returns_true():
    cl = data_provider.get_app_changelist()
    assert is_sorted_by_module(
        cl.list_key, data_provider.get_change_data('/app/src/main/res/values/strings.xml')
    )


def test_is_sorted_by_module_app_cl_src_file_returns_true():
    cl = data_provider.get_app_changelist()
    assert is_sorted_by_module(
        cl.list_key, data_provider.get_change_data('/app/src/main/java/app/Main.java')
    )


def test_is_sorted_by_module_build_updates_cl_returns_true():
    cl = data_provider.get_build_updates_changelist()
    for file in cl.changes:
        assert is_sorted_by_module(cl.list_key, file)


def test_is_sorted_by_module_build_updates_cl_gradle_properties_returns_true():
    cl = data_provider.get_build_updates_changelist()
    assert is_sorted_by_module(
        cl.list_key, data_provider.get_gradle_properties_change_data()
    )


def test_is_sorted_by_module_github_cl():
    cl = data_provider.get_github_changelist()
    assert is_sorted_by_module(
        cl.list_key, data_provider.get_change_data('/.github/workflow/test.yml')
    )


def test_is_sorted_by_module_github_cl_dependabot_returns_true():
    cl = data_provider.get_github_changelist()
    assert is_sorted_by_module(
        cl.list_key, data_provider.get_change_data('/.github/dependabot.yml')
    )



"""Testing File Validation Methods.
"""
import pytest
from pathlib import Path

from changelist_sort.input.file_validation import validate_input_file


@pytest.mark.parametrize(
    "test_input,expect",
    [
        ("file_name", "file_data"),
        ("file_name12", "file_data"),
    ]
)
def test_validate_input_file_returns_data(test_input, expect):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: "file_data")
        assert validate_input_file(test_input) == expect


def test_validate_input_file_does_not_exist_raises_exit():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        try:
            validate_input_file("file_name")
            assert False
        except SystemExit:
            assert True


def test_validate_input_file_is_empty_raises_exit():
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: "")
        try:
            validate_input_file("file_name")
            assert False
        except SystemExit:
            assert True


@pytest.mark.parametrize(
    "test_input,expect",
    [
        ("file_name", "file_data"),
        ("file_name12", "file_data"),
    ]
)
def test_validate_input_file_is_empty_returns_none(test_input, expect):
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: expect)
        assert validate_input_file(test_input) == expect



"""Defines and Validates Argument Syntax.

Encapsulates Argument Parser.

Returns Argument Data, the args provided by the User.
"""
from argparse import ArgumentParser
from sys import exit
from typing import Optional

from .argument_data import ArgumentData
from .string_validation import validate_name


def parse_arguments(args: Optional[list[str]] = None) -> ArgumentData:
    """
    Parse command line arguments.

    Parameters:
    - args: A list of argument strings.

    Returns:
    ArgumentData : Container for Valid Argument Data.
    """
    if args is None or len(args) == 0:
        return ArgumentData(None)
    # Initialize the Parser and Parse Immediately
    try:
        parsed_args = _define_arguments().parse_args(args)
    except SystemExit:
        exit("Unable to Parse Arguments.")
    return _validate_arguments(parsed_args)


def _validate_arguments(
    parsed_args,
) -> ArgumentData:
    """
    Checks the values received from the ArgParser.
        Uses Validate Name method from StringValidation.

    Parameters:
    - parsed_args : The object returned by ArgumentParser.

    Returns:
    ArgumentData - A DataClass of syntactically correct arguments.
    """
    workspace_path = parsed_args.workspace
    if workspace_path is not None:
        if not validate_name(workspace_path):
            exit("The Workspace Path argument was invalid.")
    return ArgumentData(
        workspace_path=workspace_path,
        developer_sort=parsed_args.developer_sort,
        sourceset_sort=parsed_args.sourceset_sort,
    )


def _define_arguments() -> ArgumentParser:
    """
    Initializes and Defines Argument Parser.
       - Sets Required/Optional Arguments and Flags.

    Returns:
    argparse.ArgumentParser - An instance with all supported Arguments.
    """
    parser = ArgumentParser(
        description="Changelist Sort",
    )
    # Optional Arguments
    parser.add_argument(
        '--workspace',
        type=str,
        default=None,
        help='The Workspace File containing the ChangeList data. Searches current directory by default.'
    )
    parser.add_argument(
        '-d', '--developer_sort', '--developer-sort',
        action='store_true',
        default=False,
        help='A Flag indicating that Developer Sort is to be used primarily. Fallback to Module Sort.',
    )
    parser.add_argument(
        '-s', '--sourceset_sort', '--sourceset-sort',
        action='store_true',
        default=False,
        help='A Flag indicating that SourceSet Sort is to be used primarily. Fallback to Module Sort.',
    )
    return parser



"""A Collection of String Operations.
"""


def capitalize_words(sentence: str) -> str:
    """
    Uppercase the first letter of every word in the sentence.
    """
    return ' '.join(word.capitalize() for word in sentence.split())


def replace_underscores(name: str) -> str:
    """
    Replace all underscore characters with spaces.
    """
    return name.replace('_', ' ')


def split_words_on_capitals(input_str: str) -> str:
    """
    Split a string into words on capital letters.
    """
    return ''.join(' ' + c if c.isupper() else c for c in input_str)



"""The Data Class for a ChangeList.
"""
from dataclasses import dataclass, field

from changelist_sort import list_key
from changelist_sort.change_data import ChangeData
from changelist_sort.list_key import ListKey


@dataclass(frozen=True)
class ChangelistData:
    """
    The complete Data class representing a ChangeList.
    
    Properties:
    - id (str): The unique id of the changelist.
    - name (str): The name of the changelist.
    - changes (list[ChangeData]): The list of file changes in the changelist.
    - comment (str): The comment associated with the changelist.
    - is_default (bool): Whether this is the active changelist.

    Post Init Properties:
    - simple_name (str): A simplified string derived from the name property.
    """
    id: str
    name: str
    changes: list[ChangeData] = field(default_factory=lambda: [])
    comment: str = ""
    is_default: bool = False
    
    list_key: ListKey = field(init=False)

    def __post_init__(self):
        object.__setattr__(self, 'list_key', list_key.compute_key(self.name))



""" Testing the WorkspaceTree Class.
"""
from xml.etree.ElementTree import fromstring
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.workspace.workspace_tree import WorkspaceTree
from test.data_provider import get_multi_changelist_xml
from test.data_provider import get_no_changelist_xml, get_simple_changelist_xml


def get_simple_ws_tree():
    return WorkspaceTree(fromstring(get_simple_changelist_xml()))

def get_multi_ws_tree():
    return WorkspaceTree(fromstring(get_multi_changelist_xml()))

def get_no_cl_ws_tree():
    return WorkspaceTree(fromstring(get_no_changelist_xml()))


def test_extract_list_elements_simple_returns_list():
    result = get_simple_ws_tree().extract_list_elements()
    assert len(result) == 1
    cl = result[0]
    assert cl.name == 'Simple'
    assert cl.comment == 'Main Program Files'
    assert len(cl.changes) == 1
    file = cl.changes[0]
    assert file.before_path == '/main.py'
    assert not file.before_dir
    assert file.after_path == '/main.py'
    assert not file.after_dir


def test_extract_list_elements_multi_returns_list():
    result = get_multi_ws_tree().extract_list_elements()
    assert len(result) == 2
    # First Changelist
    cl_0 = result[0]
    assert cl_0.name == 'Main'
    assert len(cl_0.changes) == 2
    # Second Changelist
    cl_1 = result[1]
    assert cl_1.name == 'Test'
    assert len(cl_1.changes) == 1


def test_extract_list_elements_no_cl_returns_empty_list():
    ws_tree = get_no_cl_ws_tree()
    try:
        ws_tree.extract_list_elements()
        assert False
    except SystemExit:
        assert True


def test_replace_changelists_simple_with_empty():
    ws_tree = get_simple_ws_tree()
    ws_tree.replace_changelists([])
    # Get Elements
    result = ws_tree.extract_list_elements()
    assert len(result) == 0


def test_replace_changelists_simple_with_multi():
    ws_tree = get_simple_ws_tree()
    ws_tree.replace_changelists([
        ChangelistData(
            id='af84ea1b',
            name='Main',
            changes=[],
        ),
        ChangelistData(
            id='9f60fda2',
            name='Test',
            changes=[],
        ),
    ])
    # Get Elements
    result = ws_tree.extract_list_elements()
    assert len(result) == 2


def test_replace_changelists_no_cl_manager_raises_exit():
    ws_tree = WorkspaceTree(fromstring(get_no_changelist_xml()))
    try:
        ws_tree.replace_changelists([])
        raised_exit = False
    except SystemExit:
        raised_exit = True
    assert raised_exit


def test_get_root_no_cl_manager_returns_root():
    ws_tree = WorkspaceTree(fromstring(get_no_changelist_xml()))
    xml_root = ws_tree.get_root()
    assert xml_root.getroot().tag == 'project'



""" Testing Change Sort Module Methods
"""
from test import data_provider

from changelist_sort.change_data import ChangeData
from changelist_sort.sorting.file_sort import get_module_name, get_module_type
from changelist_sort.sorting.module_type import ModuleType


def test_get_module_name_empty_returns_none():
    assert get_module_name(ChangeData()) is None


def test_get_module_name_requirements_file_returns_module():
    assert 'root' == get_module_name(
        data_provider.get_change_data('requirements.txt')
    )


def test_get_module_name_setup_file_returns_module():
    assert 'root' == get_module_name(
        data_provider.get_change_data('setup.py')
    )


def test_get_module_name_src_file_returns_module():
    assert 'module' == get_module_name(
        data_provider.get_module_src_change_data()
    )
    

def test_get_module_name_test_file_returns_module():
    assert 'module' == get_module_name(
        data_provider.get_module_test_change_data()
    )


def test_get_module_name_app_build_file_returns_app():
    assert 'gradle' == get_module_name(
        data_provider.get_app_gradle_build_change_data()
    )


def test_get_module_name_root_build_file_returns_root():
    assert 'gradle' == get_module_name(
        data_provider.get_root_gradle_build_change_data()
    )


def test_get_module_name_github_workflows_returns_github():
    assert 'github' == get_module_name(
        data_provider.get_github_workflows_change_data()
    )


def test_get_module_name_github_dependabot_returns_github():
    assert 'github' == get_module_name(
        data_provider.get_github_dependabot_change_data()
    )


def test_get_module_type_root_gradle_returns_gradle():
    assert ModuleType.GRADLE == get_module_type(
        data_provider.get_root_gradle_build_change_data()
    )


def test_get_module_type_root_gradle_kts_returns_gradle():
    assert ModuleType.GRADLE == get_module_type(data_provider.get_change_data(
        '/build.gradle.kts'
    ))


def test_get_module_type_gradle_properties_returns_gradle():
    assert ModuleType.GRADLE == get_module_type(data_provider.get_change_data(
        data_provider.GRADLE_PROPERTIES_PATH
    ))


def test_get_module_type_app_build_file_returns_module():
    assert ModuleType.GRADLE == get_module_type(
        data_provider.get_app_gradle_build_change_data()
    )


def test_get_module_type_app_src_file_returns_module():
    assert ModuleType.MODULE == get_module_type(data_provider.get_change_data(
        '/app/src/main/java/com/example/app/Main.java'
    ))


def test_get_module_type_app_test_file_returns_module():
    assert ModuleType.MODULE == get_module_type(data_provider.get_change_data(
        '/app/src/test/java/com/example/app/Main.java'
    ))


def test_get_module_type_app_res_file_returns_module():
    assert ModuleType.MODULE == get_module_type(data_provider.get_change_data(
        '/app/src/main/res/values/strings.xml'
    ))


def test_get_module_type_github_workspace_returns_hidden():
    assert ModuleType.HIDDEN == get_module_type(data_provider.get_change_data(
        data_provider.GITHUB_WORKFLOW_PATH
    ))



""" Sorting Package.
"""
from typing import Callable

from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.list_key import ListKey
from changelist_sort.sorting.sort_mode import SortMode
from changelist_sort.sorting import developer_sort, module_sort, source_set_sort
from changelist_sort.sorting.list_sort import split_changelist


def sort(
    initial_list: list[ChangelistData],
    sort_mode: SortMode,
) -> list[ChangelistData]:
    """
    Processes InputData.

    Parameters:
    - initial_list (list[ChangelistData]): The list of Changelists to be sorted.
    - sort_mode (SortMode): The SortMode determining which sort rules to apply.

    Returns:
    str - The desired output.
    """
    unsorted_files = []
    cl_map = ChangelistMap()
    for cl in initial_list:
        if not cl_map.insert(cl):
            _handle_map_insertion_error()
        # Split CL based on SortMode criteria
        unsorted_files.extend(
            split_changelist(cl, _get_is_sorted_callable(sort_mode))
        )
    # This Callable depends on SortMode.
    #  It determines map keys, and executes map insertions.
    sort_result_string = _process_file_sort_results(list(map(
        _create_sorting_callable(cl_map, sort_mode),
        unsorted_files
    )))
    print(sort_result_string)
    return cl_map.get_lists()


def _create_sorting_callable(
    changelist_map: ChangelistMap,
    sort_mode: SortMode,
) -> Callable[[ChangeData], bool]:
    """
    Create a Callable that sorts ChangeData passed to it.
    """
    if sort_mode == SortMode.MODULE:
        return lambda x: module_sort.sort_file_by_module(changelist_map, x)
    if sort_mode == SortMode.DEVELOPER:
        return lambda x: developer_sort.sort_file_by_developer(changelist_map, x)
    if sort_mode == SortMode.SOURCESET:
        return lambda x: source_set_sort.sort_by_source_set(changelist_map, x)
    else:
        exit("SortMode not Implemented")


def _get_is_sorted_callable(
    sort_mode: SortMode
) -> Callable[[ListKey, ChangeData], bool]:
    """
    Obtain a Callable that determines whether a ChangeData is sorted.
    """
    if sort_mode == SortMode.MODULE:
        return module_sort.is_sorted_by_module
    if sort_mode == SortMode.DEVELOPER:
        return developer_sort.is_sorted_by_developer
    if sort_mode == SortMode.SOURCESET:
        return source_set_sort.is_sorted_by_source_set
    else:
        exit("SortMode not Implemented")


def _handle_map_insertion_error(
    cl_map: ChangelistMap,
    failure_cl: ChangelistData,
):
    """
    Using the given parameters, produce an error message and exit.

    Raises:
    SystemExit - containing error information.
    """
    if (existing_cl := cl_map.search(failure_cl.simple_name)) is not None:
        exit(f"Failed to Insert Changelist(name={failure_cl.name}) due to simplename key conflict with Changelist(name={existing_cl.name}).")
    elif cl_map.contains_id(failure_cl.id):
        exit(f"Failed to Insert Changelist(name={failure_cl.name}) due to id conflict (id={failure_cl.id}).")
    else:
        exit(f"Failed to Insert Changelist(name={failure_cl.name}) for unknown reason (neither key nor id conflict has occurred).")


def _process_file_sort_results(results: list[bool]) -> str:
    """
    Process and Format a string to print.
    """
    correct = sum(results)
    total = len(results)
    return f"Sorted {correct} of {total} unsorted files."



""" Workspace Package.
"""
from pathlib import Path
from xml.etree.ElementTree import ParseError, fromstring

from changelist_sort.input.input_data import InputData
from changelist_sort.workspace.workspace_tree import WorkspaceTree


def get_workspace_tree(
    input_data: InputData
) -> WorkspaceTree:
    """
    Parse the Workspace XML into a Workspace Tree.

    Parameters:
    - input_data (InputData)
    """
    try:
        xml_root = fromstring(input_data.workspace_xml)  
    except ParseError:
        exit(f'Failed to Parse Workspace XML at {input_data.workspace_path}')
    return WorkspaceTree(xml_root)


def write_workspace_tree(
    ws_tree: WorkspaceTree,
    path: Path,
) -> bool:
    """
    Write the Workspace Tree as XML to the given Path.
    """
    ws_tree.get_root().write(
        file_or_filename=path,
        encoding='utf-8',
        xml_declaration=True,
        method='xml',
    )



""" Sorting By Module.
    An option of SortMode.
"""
from changelist_sort.list_key import ListKey
from changelist_sort.sorting import file_sort
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.string_operations import capitalize_words
from changelist_sort.string_operations import replace_underscores


MODULE_ROOT_CL_TUPLE = (
    ListKey('projectroot', 'Project Root'),
    ListKey('root', 'Project Root'),
)

MODULE_GRADLE_CL_TUPLE = (
    ListKey('buildupdates', 'Build Updates'),
    ListKey('gradle', 'Build Updates'),
)


def get_module_keys(module_type: ModuleType) -> tuple[str, ...]:
    """
    Obtain a tuple containing the Keys for the Changelists.

    Parameters:
    - module_type (ModuleType): The type of Module.

    Returns
    tuple[str] - a tuple containing keys to lists of a given type of module.
    """
    if module_type == ModuleType.ROOT:
        return tuple(x.key for x in MODULE_ROOT_CL_TUPLE)
    elif module_type == ModuleType.GRADLE:
        return tuple(x.key for x in MODULE_GRADLE_CL_TUPLE)
    else:
        return tuple()


def sort_file_by_module(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    """
    Sort files into Changelists by Module.

    Parameters:
    - cl_map (ChangelistMap): The Map of Changelists to sort into.
    - file (ChangeData): The File change data to be sorted.

    Returns:
    bool - True when the operation succeeds.
    """
    if (module_type := file_sort.get_module_type(file)) is None:
        return False
    if module_type == ModuleType.ROOT:
        return _sort_root_module(cl_map, file)
    elif module_type == ModuleType.GRADLE:
        return _sort_gradle_module(cl_map, file)
    else:
        return _sort_module(cl_map, file)


def is_sorted_by_module(
    cl_key: ListKey,
    file: ChangeData,
) -> bool:
    """
    Determine whether this file belongs in the given Changelist.
        Applies Special Module, and Directory equivalencies.

    Parameters:
    - changelist_name (str): The name of the Changelist, used to determine if the file is sorted.
    - file (ChangeData): The File to compare against the Changelist name.

    Returns:
    bool - Whether this file belongs in this Changelist according to Module sort logic.
    """
    if (module_type := file_sort.get_module_type(file)) == ModuleType.ROOT:
        if cl_key.key.startswith(get_module_keys(ModuleType.ROOT)):
            return True
        # File Extension Checks
        return file.file_ext.endswith(
            file_sort._GRADLE_FILE_SUFFIXES
        ) and cl_key.key.startswith(
            get_module_keys(ModuleType.GRADLE)
        )
    if module_type == ModuleType.GRADLE:
        return cl_key.key.startswith(get_module_keys(ModuleType.GRADLE))
    # Starts with or Equals
    return cl_key.key.startswith(
        file_sort.get_module_name(file)
    )


def _sort_root_module(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    # Find CL
    for module_cl in MODULE_ROOT_CL_TUPLE:
        if (existing_cl := cl_map.search(module_cl.key)) is not None:
            existing_cl.changes.append(file)
            return True
    # Create New ChangeList and Append File
    cl_map.create_changelist(
        MODULE_ROOT_CL_TUPLE[0],
    ).changes.append(file)
    return True


def _sort_gradle_module(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    # Find CL
    for module_cl in MODULE_GRADLE_CL_TUPLE:
        if (existing_cl := cl_map.search(module_cl.key)) is not None:
            existing_cl.changes.append(file)
            return True
    # Create New ChangeList and Append File
    cl_map.create_changelist(
        MODULE_GRADLE_CL_TUPLE[0]
    ).changes.append(file)
    return True


def _sort_module(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    # Get Module Name from File Path data
    if (file_module := file_sort.get_module_name(file)) is None or len(file_module) == 0:
        return False
    if (cl := cl_map.search(file_module)) is not None:
        cl.changes.append(file)
        return True
    # Create Changelist and Append File
    cl_map.create_changelist(
        capitalize_words(replace_underscores(file_module))
    ).changes.append(file)
    return True



"""File Validation Methods.
    These Methods all raise SystemExit exceptions.
"""
from pathlib import Path
from sys import exit

from .string_validation import validate_name


def validate_input_file(file_name: str) -> str:
    """
    Read the Input File, Validate (non-blank) data, and return Input str.

    Parameters:
    - file_name (str): The Name of the Input File.

    Returns:
    str - The String Contents of the Input File.

    Raises:
    SystemExit - If the File does not exist, or is empty or blank, or read failed.
    """
    file_path = Path(file_name)
    if not file_path.exists():
        exit("The Input File does not Exist.")
    try:
        data = file_path.read_text()
    except IOError:
        exit("Failed to Read from File.")
    if validate_name(data):
        return data
    exit("Input was Empty")



""" Testing Workspace Tree.
"""
from xml.etree.ElementTree import fromstring

from test import data_provider
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.workspace.workspace_tree import _convert_bool, _extract_change_data, _filter_project_dir, _find_changelist_manager, _write_list_element


def test_find_changelist_manager_no_changelist_returns_none():
    xml_root = fromstring(data_provider.get_no_changelist_xml())
    assert _find_changelist_manager(xml_root) is None


def test_find_changelist_manager_simple_changelist_returns_element():
    xml_root = fromstring(data_provider.get_simple_changelist_xml())
    element = _find_changelist_manager(xml_root)
    cl_elements = list(element.iter())
    assert len(cl_elements) == 3
    assert cl_elements[0].attrib['name'] == 'ChangeListManager'
    assert cl_elements[1].tag == 'list'
    assert cl_elements[2].tag == 'change'


def test_find_changelist_manager_multi_changelist_returns_element():
    xml_root = fromstring(data_provider.get_multi_changelist_xml())
    element = _find_changelist_manager(xml_root)
    cl_elements = list(element.iter())
    assert len(cl_elements) == 6
    cl_main = cl_elements[1]
    assert cl_main.tag == 'list'
    assert cl_elements[2].tag == 'change'
    assert cl_elements[3].tag == 'change'
    cl_test = cl_elements[4]
    assert cl_test.tag == 'list'
    assert cl_elements[5].tag == 'change'


def test_find_changelist_manager__returns_none():
    xml_root = fromstring(data_provider.get_invalid_component_xml())
    element = _find_changelist_manager(xml_root)
    assert element is not None


def test_extract_change_data_simple_():
    xml_root = fromstring(data_provider.get_simple_changelist_xml())
    element = _find_changelist_manager(xml_root)
    cl_elements = list(element.iter())
    # Extract Change Data is called on a List element
    c_data_list = _extract_change_data(cl_elements[1])
    assert len(c_data_list) == 1
    c_data0 = c_data_list[0]
    assert c_data0.before_path == '/main.py'
    assert not c_data0.before_dir
    assert c_data0.after_path == '/main.py'
    assert not c_data0.after_dir


def test_extract_change_data_empty_element_returns_empty_list():
    xml_root = fromstring(data_provider.get_simple_changelist_xml())
    element = _find_changelist_manager(xml_root)
    cl_elements = list(element.findall('list'))
    assert len(cl_elements) == 1
    # Remove sub_element of the List element
    cl_elements[0].remove(cl_elements[0][0])
    assert len(cl_elements[0]) == 0
    assert len(_extract_change_data(cl_elements[0])) == 0


def test_write_list_element_simple_returns_element():
    cl_data = ChangelistData(
        id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88",
        name="Simple",
        changes=[],
        comment="Main Program Files",
        is_default=False,
    )
    cl_element = _write_list_element(cl_data)
    #
    assert len(cl_element.findall('change')) == 0
    assert cl_element.attrib['id'] == '9f60fda2-421e-4a4b-bd0f-4c8f83a47c88'
    assert cl_element.attrib['name'] == 'Simple'
    assert cl_element.attrib['comment'] == 'Main Program Files'
    # Default is not included when false
    assert 'default' not in cl_element.attrib


def test_write_list_element_default_returns_element():
    cl_data = ChangelistData(
        id="9f60fda2-421e-4a4b-bd0f-4c8f83a47c88",
        name="Simple",
        changes=[],
        comment="Main Program Files",
        is_default=True,
    )
    cl_element = _write_list_element(cl_data)
    #
    assert len(cl_element.findall('change')) == 0
    assert cl_element.attrib['id'] == '9f60fda2-421e-4a4b-bd0f-4c8f83a47c88'
    assert cl_element.attrib['name'] == 'Simple'
    assert cl_element.attrib['comment'] == 'Main Program Files'
    assert cl_element.attrib['default'] == 'true'


def test_write_list_element_changes_returns_element():
    cl_data = data_provider.get_build_updates_changelist(
        changes=[data_provider.get_root_gradle_build_change_data()]
    )
    print(', '.join(x.sort_path for x in cl_data.changes))
    assert len(cl_data.changes) == 1
    cl_element = _write_list_element(cl_data)
    #
    assert len(cl_element.findall('change')) == 1
    assert cl_element.attrib['id'] == cl_data.id
    assert cl_element.attrib['name'] == cl_data.name
    assert cl_element.attrib['comment'] == cl_data.comment
    assert 'default' not in cl_element.attrib


def test_filter_project_dir_none_returns_none():
    assert _filter_project_dir(None) is None


def test_filter_project_dir_normal_returns_shortened_path():
    normal_input = '$PROJECT_DIR$/main.py'
    assert _filter_project_dir(normal_input) == '/main.py'


def test_filter_project_dir_without_returns_same_path():
    test_input = '/main.py'
    assert _filter_project_dir(test_input) == '/main.py'


def test_convert_bool_none_returns_none():
    assert _convert_bool(None) is None


def test_convert_bool_false_returns_false():
    assert not _convert_bool('false')


def test_convert_bool_true_returns_true():
    assert _convert_bool('true')



"""Testing Argument Parser Methods.
"""
from changelist_sort.input.argument_parser import parse_arguments


def test_parse_arguments_empty_list_returns_none():
    result = parse_arguments()
    assert result.workspace_path is None


def test_parse_arguments_empty_str_returns_none():
    result = parse_arguments('')
    assert result.workspace_path is None


def test_parse_arguments_change_list_main_empty_workspace_arg():
    try:
        parse_arguments(['--workspace', ''])
        assert False
    except SystemExit:
        assert True


def test_parse_arguments_change_list_main():
    result = parse_arguments(['--workspace', 'workspace.xml'])
    assert result.workspace_path == 'workspace.xml'


def test_parse_arguments_developer_sort_short_flag():
    result = parse_arguments(['-d'])
    assert result.workspace_path is None
    assert result.developer_sort
    assert not result.sourceset_sort


def test_parse_arguments_developer_sort_long():
    result = parse_arguments(['--developer-sort'])
    assert result.workspace_path is None
    assert result.developer_sort
    assert not result.sourceset_sort


def test_parse_arguments_developer_sort_long2():
    result = parse_arguments(['--developer_sort'])
    assert result.workspace_path is None
    assert result.developer_sort
    assert not result.sourceset_sort


def test_parse_arguments_sourceset_short_flag():
    result = parse_arguments(['-s'])
    assert result.workspace_path is None
    assert not result.developer_sort
    assert result.sourceset_sort


def test_parse_arguments_sourceset_long():
    result = parse_arguments(['--sourceset-sort'])
    assert result.workspace_path is None
    assert not result.developer_sort
    assert result.sourceset_sort


def test_parse_arguments_sourceset_long2():
    result = parse_arguments(['--sourceset_sort'])
    assert result.workspace_path is None
    assert not result.developer_sort
    assert result.sourceset_sort



"""The Arguments Received from the Command Line Input.

This DataClass is created after the argument syntax is validated.

Syntax Validation:
-
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class ArgumentData:
    """
    The syntactically valid arguments recevied by the Program.

    Fields:
    - workspace_path (str): The path to the workspace file.
    - developer_sort (bool):
    - sourceset_sort (bool):
    """
    workspace_path: str | None
    developer_sort: bool = False
    sourceset_sort: bool = False



"""Valid Input Data Class.
"""
from dataclasses import dataclass
from pathlib import Path

from changelist_sort.sorting.sort_mode import SortMode


@dataclass(frozen=True)
class InputData:
    """A Data Class Containing Program Input.

    Fields:
    - workspace_xml (str): The contents of the Workspace XML file.
    - workspace_path (Path): The Path to the Workspace File.
    - sort_mode (SortMode): The selected Sorting Mode enum value.
    """
    workspace_xml: str
    workspace_path: Path
    sort_mode: SortMode = SortMode.MODULE



""" Testing Main.
"""
from pathlib import Path
from xml.etree.ElementTree import ElementTree
import pytest
from test import data_provider


def test_main_():
    with pytest.MonkeyPatch().context() as ctx:
        #ctx.setenv()
        import sys
        original_argv = sys.argv
        sys.argv = ['changelist_sort', '--workspace', '/mockfile.xml']
        # Now Import
        from changelist_sort.__main__ import main
        # Mock File Input
        ctx.setattr(Path, 'exists', lambda _: True)
        ctx.setattr(Path, 'read_text', lambda _: data_provider.get_simple_changelist_xml())
        # Mock File Write
        def element_tree_write(
            self: ElementTree, file_or_filename, encoding='utf-8',
            xml_declaration=True, method='xml',
            default_namespace=None,
            short_empty_elements=True,
        ):
            global TAG
            global VERSION
            elem = self.getroot()
            TAG = elem.tag
            VERSION = elem.attrib['version']
        ctx.setattr(ElementTree, 'write', element_tree_write)
        # Run Main
        main()
        assert TAG == 'project'
        assert VERSION == '4'
        sys.argv = original_argv


""" Testing Input Package Init Module Methods.
"""
import pytest
from pathlib import Path

from changelist_sort.input import validate_input
from changelist_sort.sorting.sort_mode import SortMode
from test import data_provider


def test_validate_input_no_args_ws_file_does_not_exist_raises_exit():
    test_input = []
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        try:
            validate_input(test_input)
            assert False
        except SystemExit:
            assert True


def test_validate_input_no_args_ws_file_is_empty_raises_exit():
    test_input = []
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: '')
        try:
            validate_input(test_input)
            assert False
        except SystemExit:
            assert True


def test_validate_input_no_args_ws_file_has_no_cl_():
    test_input = []
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_no_changelist_xml())
        result = validate_input(test_input)
        assert len(result.workspace_xml) > 1
        


def test_validate_input_no_args_ws_file_simple_cl_():
    test_input = []
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_simple_changelist_xml())
        result = validate_input(test_input)
        assert len(result.workspace_xml) > 1


def test_validate_input_ws_path_arg_is_empty_raises_exit():
    test_input = ['--workspace', '']
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_no_changelist_xml())
        try:
            validate_input(test_input)
            assert False
        except SystemExit:
            assert True


def test_validate_input_ws_path_arg_is_missing_raises_exit():
    test_input = ['--workspace']
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_no_changelist_xml())
        try:
            validate_input(test_input)
            assert False
        except SystemExit:
            assert True


def test_validate_input_ws_path_arg_does_not_exist_raises_exit():
    test_input = ['--workspace', '/file.xml']
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: False)
        try:
            validate_input(test_input)
            assert False
        except SystemExit:
            assert True


def test_validate_input_developer_sort():
    test_input = ['-d']
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_no_changelist_xml())
        result = validate_input(test_input)
        assert result.workspace_path == Path('.idea/workspace.xml')
        assert result.sort_mode == SortMode.DEVELOPER


def test_validate_input_sourceset_sort():
    test_input = ['-s']
    with pytest.MonkeyPatch().context() as c:
        c.setattr(Path, 'exists', lambda _: True)
        c.setattr(Path, 'read_text', lambda _: data_provider.get_no_changelist_xml())
        result = validate_input(test_input)
        assert result.workspace_path == Path('.idea/workspace.xml')
        assert result.sort_mode == SortMode.SOURCESET





"""Testing String Operations.
"""
from changelist_sort.string_operations import capitalize_words, split_words_on_capitals


def test_captialize_word_app():
    assert capitalize_words('app') == 'App'


def test_captialize_word_gradle():
    assert capitalize_words('gradle') == 'Gradle'


def test_captialize_word_build_updates():
    assert capitalize_words('build updates') == 'Build Updates'


def test_captialize_word_root_project():
    assert capitalize_words('root project') == 'Root Project'


def test_split_words_on_capitals_main_returns_main():
    assert 'main' == split_words_on_capitals('main')


def test_split_words_on_capitals_android_test_returns_android_test():
    assert 'android test' == split_words_on_capitals('android test')


def test_split_words_on_capitals_test_fixtures_returns_test_fixtures():
    assert 'test fixtures' == split_words_on_capitals('test fixtures')



""" Testing String Validation Methods
"""

from changelist_sort.input.string_validation import validate_name


def test_validate_name_empty_returns_false():
    assert not validate_name('')


def test_validate_name_space_returns_false():
    assert not validate_name(' ')


def test_validate_name_simple_returns_true():
    assert validate_name('a')


def test_validate_name_non_string_returns_false():
    assert not validate_name(2)



""" List Keys Definitions.
"""
from dataclasses import dataclass


@dataclass(frozen=True)
class ListKey:
    """
    A Changelist Name and Key.

    Parameters:
    - key (str): The key used to quickly search for the changelist.
    - changelist_name (str): The full user defined name for the changelist.
    """
    key: str
    changelist_name: str


def compute_key(cl_name: str) -> ListKey:
    """
    Compute a Key to use for a given Changelist Name.
    - computation is a series of reduction operations

    Parameters:
    - cl_name (str): The Changelist Name to be included in the ListKey.

    Returns:
    ListKey - A ListKey instance contianing the Key and Changelist Name.
    """
    if len(cl_name) == 0:
        return ListKey('', '')
    translator = str.maketrans('', '', ' :/\\')
    words = cl_name.translate(translator).split()
    return ListKey(
        key=''.join(w.lower() for w in words),
        changelist_name=cl_name
    )



"""The Input Package for Changelist Sort
"""
from pathlib import Path

from changelist_sort.input.argument_data import ArgumentData
from changelist_sort.input.argument_parser import parse_arguments
from changelist_sort.input.file_validation import validate_input_file
from changelist_sort.input.input_data import InputData
from changelist_sort.sorting.sort_mode import SortMode


def validate_input(args_list: list[str]) -> InputData:
    """
    Validate the arguments and gather program input into InputData object.
    - Parses command line strings into ArgumentData object.
    - Finds Workspace file and reads it to a string.

    Returns:
    InputData - container for the program input.
    """
    arg_data = parse_arguments(args_list)
    ws_path = _find_workspace_file(arg_data)

    return InputData(
        workspace_xml=validate_input_file(ws_path),
        workspace_path=ws_path,
        sort_mode=_determine_sort_mode(arg_data),
    )


def _find_workspace_file(arg_data: ArgumentData) -> Path:
    """
    Generate a Path to the Workspace File.
    - If Path is given by CLI arguments, use it.
    - Otherwise, assume current directory is project root.

    Returns:
    Path - A Path object. Note, the actual file may not exist.
    """
    if arg_data.workspace_path is None:
        return Path('.') / '.idea' / 'workspace.xml'
    return Path(arg_data.workspace_path)


def _determine_sort_mode(arg_data: ArgumentData) -> SortMode:
    """
    Check the Argument Data flags to determine which SortMode to use.
    """
    if arg_data.developer_sort:
        return SortMode.DEVELOPER
    if arg_data.sourceset_sort:
        return SortMode.SOURCESET
    return SortMode.MODULE



""" Sorting By Gradle Module Source Set.

Builds on the Module Sort, with additional subcategories for Gradle Source Sets.
"""
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.list_key import ListKey
from changelist_sort.sorting import file_sort, module_sort
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.string_operations import capitalize_words, replace_underscores
from changelist_sort.string_operations import split_words_on_capitals


def sort_by_source_set(
    cl_map: ChangelistMap,
    file: ChangeData,
) -> bool:
    """
    Sort files into Changelists by Source Set.
    - Is Based on Module Sort. Should defer to Module Sort where possible.
    - Matches 

    Parameters:
    - cl_map (ChangelistMap): The Map of Changelists to sort into.
    - file (ChangeData): The File change data to be sorted.

    Returns:
    bool - True when the operation succeeds.
    """
    if file_sort.get_module_type(file) == ModuleType.MODULE:
        file_module = file_sort.get_module_name(file)
        # Get SourceSet Name
        if (source_set_name := _get_source_set_name(file)) is not None:
            # Merge Module and SourceSet into a Sorting Key
            sort_key = f"{file_module}{source_set_name.lower()}"
            if (cl := cl_map.search(sort_key)) is not None:
                cl.changes.append(file)
                return True
            # Create a new Changelist for this SourceSet
            cl_map.create_changelist(
                capitalize_words(replace_underscores(
                    f"{file_module} {split_words_on_capitals(source_set_name)}"
                ))
            ).changes.append(file)
            return True
    # Fallback to ModuleSort
    return module_sort.sort_file_by_module(cl_map, file)


def is_sorted_by_source_set(
    cl_key: ListKey,
    file: ChangeData,
) -> bool:
    """
    Determine whether this file belongs in the given Changelist.
        Applies Special Module, and Directory equivalencies.

    Parameters:
    - changelist_name (str): The name of the Changelist, used to determine if the file is sorted.
    - file (ChangeData): The File to compare against the Changelist name.

    Returns:
    bool - Whether this file belongs in this Changelist according to Module sort logic.
    """
    if file_sort.get_module_type(file) == ModuleType.MODULE:
        file_module = file_sort.get_module_name(file)
        # Get SourceSet Name
        if (source_set_name := _get_source_set_name(file)) is not None:
            if cl_key.key.startswith(f"{file_module}{source_set_name.lower()}"):
                return True
            return False
    return module_sort.is_sorted_by_module(cl_key, file)


def _get_source_set_name(file: ChangeData) -> str | None:
    """
    Determine the Source Set of the File, if possible.

    Parameters:
    - file (ChangeData): The File data to obtain the Source Set name from.

    Returns:
    str | None - The SourceSet name, or None if not applicable.
    """
    if (path := file.sort_path) is None:
        return None
    # Find the Source Dir
    if (src_start_idx := path.find('/src/') + 5) < 5:
        return None
    # Find the next dir slash
    if (src_end_idx := path.find('/', src_start_idx)) <= src_start_idx:
        return None
    # Cut path string
    return path[src_start_idx:src_end_idx]



""" Changelist Sorting Methods.
"""
from typing import Callable

from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.list_key import ListKey


def split_changelist(
    changelist: ChangelistData,
    is_sorted: Callable[[ListKey, ChangeData], bool],
) -> list[ChangeData]:
    """
    Split the Changelist by checking that all changes are sorted.
        Removes each element that is returned from the changelist.

    Parameters:
    - changelist (ChangelistData): The Changelist to split based on sorting function.
    - is_sorted (callable[bool]): A Function that determines whether a Change is Sorted.

    Returns:
    list[ChangeData] - The List of ChangeData that are Unsorted, now removed from this changelist.
    """
    unsorted_files = []
    for index in range(len(changelist.changes) - 1, -1, -1):
        if not is_sorted(
            changelist.list_key,
            changelist.changes[index]
        ):
            unsorted_files.append(
                changelist.changes.pop(index)
            )
    return unsorted_files



#!/usr/bin/python
from sys import argv, path


def main():
    from changelist_sort.input import validate_input
    from changelist_sort import sort_changelists
    input_data = validate_input(argv[1:])
    sort_changelists(input_data)
    #print(output_data)


if __name__ == "__main__":
    from pathlib import Path
    # Get the directory of the current file (__file__ is the path to the script being executed)
    current_directory = Path(__file__).resolve().parent.parent
    # Add the directory to sys.path
    path.append(str(current_directory))
    main()



""" The Workspace Element Tree Methods.
"""
from xml.etree.ElementTree import Element, ElementTree, indent

from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.workspace.xml_reader import _read_bool_from, filter_by_tag, get_attr, get_attr_or


class WorkspaceTree:
    """
    Manages the Workspace XML Element Trees.

    Properties:
    - xml_root (Element): The XML root element.
    - changelist_manager (Element): The Changelist Manager Component Element.
    """

    def __init__(
        self,
        xml_root: Element,
    ):
        self._xml_root = xml_root
        self.changelist_manager = _find_changelist_manager(xml_root)

    def extract_list_elements(self) -> list[ChangelistData]:
        """
        Given the Changelist Manager Element, obtain the list of List Elements.

        Parameters:
        - changelist_manager (ElementTree): The ChangeList Manager XML Element.

        Returns:
        list[Element] - A List containing the Lists.
        """
        if self.changelist_manager is None:
            exit('XML File does not have a Changelist Manager.')
        return [
            ChangelistData(
                id=get_attr(cl_element, 'id'),
                name=get_attr(cl_element, 'name'),
                changes=_extract_change_data(cl_element),
                comment=get_attr_or(cl_element, 'comment', ''),
                is_default=_read_bool_from(cl_element, 'default'),
            ) for cl_element in filter_by_tag(self.changelist_manager, 'list')
        ]

    def replace_changelists(
        self,
        changelists: list[ChangelistData]
    ):
        """
        Replace the XML Tree's Changelist Manager Lists.
        """
        clm = self.changelist_manager
        if clm is None:
            exit('XML File does not have a Changelist Manager.')
        # First obtain all Option Elements
        options = list(clm.findall('option'))
        # Clear the Changelist Manager Tag
        clm.clear()
        # Need to Add Name Attribute after Clear
        clm.attrib['name'] = "ChangeListManager"
        clm.extend(_write_list_element(x) for x in changelists)
        clm.extend(options)
        indent(clm, level=1)

    def get_root(self) -> ElementTree:
        """
        Obtain the XML ElementTree Root.
        """
        return ElementTree(self._xml_root)


def _find_changelist_manager(xml_root: Element) -> Element | None:
    """
    Try to find the Changelist Manager in the XML ElementTree.
    """
    cl_manager = None
    for elem in filter_by_tag(xml_root, 'component'):
        try:
            if elem.attrib["name"] == 'ChangeListManager':
                cl_manager = elem
                break
        except KeyError:
            pass
    return cl_manager


def _extract_change_data(
    list_element: Element,
) -> list[ChangeData]:
    """
    Given a ChangeList XML Element, obtain the List of Changes.

    Parameters:
    - list_element (Element): The Element representing a Changelist.

    Returns:
    list[ChangeData] - The list of structured ChangeData.
    """
    return [
        ChangeData(
            before_path=_filter_project_dir(get_attr(change, 'beforePath')),
            before_dir=_convert_bool(get_attr(change, 'beforeDir')),
            after_path=_filter_project_dir(get_attr(change, 'afterPath')),
            after_dir=_convert_bool(get_attr(change, 'afterDir')),
        ) for change in filter(lambda x: x.tag == 'change', list_element)
    ]


def _write_list_element(
    changelist: ChangelistData,
) -> Element:
    """
    Convert a Changelist to XML Element format.

    Parameters:
    - changelist (ChangelistData): The Changelist data to format.

    Returns:
    str - A String containing the xml formatted contents of the Changelist. 
    """
    clist = Element('list')
    if changelist.is_default:
        clist.set('default', 'true')
    clist.set('id', changelist.id)
    clist.set('name', changelist.name)
    clist.set('comment', changelist.comment)
    for change in changelist.changes:
        clist.append(_write_change_data(change))
    indent(clist, level=2)
    return clist


def _write_change_data(
    data: ChangeData
) -> Element:
    """
    Write the Change Data to XML format.

    Parameters:
    - data (ChangeData): The Change Data for a specific File.

    Returns:
    str - A String containing the XML tag for this Change Data.
    """
    change = Element('change')
    if data.before_path is not None:
        change.set('beforePath', _PROJECT_DIR_VAR + data.before_path)
    if data.before_dir is not None:
        change.set('beforeDir', str(data.before_dir).lower())
    if data.after_path is not None:
        change.set('afterPath', _PROJECT_DIR_VAR + data.after_path)
    if data.after_dir is not None:
        change.set('afterDir', str(data.after_dir).lower())
    #
    indent(change, level=3)
    return change


_PROJECT_DIR_VAR = '$PROJECT_DIR$'
_PROJECT_DIR_LEN = len(_PROJECT_DIR_VAR)


def _filter_project_dir(path_str: str | None) -> str | None:
    """
    Filter the ProjectDir string at the beginning of the path.
    """
    if path_str is None:
        return None
    if path_str.startswith(_PROJECT_DIR_VAR):
        return path_str[_PROJECT_DIR_LEN:]
    return path_str


def _convert_bool(attr: str | None) -> bool | None:
    """
    Convert a String attribute to boolean.
    """
    if attr is None:
        return None
    return attr == 'true'



""" Testing Developer Changelist
"""
from test import data_provider

from changelist_sort import list_key
from changelist_sort.sorting.developer_sort import _INPUT_PACKAGE_PATTERN, _SRC_DIR_PATTERN
from changelist_sort.workspace.developer_changelist import DeveloperChangelist


def get_dc_src_input() -> DeveloperChangelist:
    return DeveloperChangelist(
        None,
        list_key.compute_key('Input Source Package'),
        (
            _SRC_DIR_PATTERN,
            _INPUT_PACKAGE_PATTERN
        )
    )


def test_check_file_src_dir_input_pattern_input_data_returns_true():
    instance = get_dc_src_input()
    assert instance.check_file(data_provider.get_change_data('/changelist_sort/input/input_data.py'))
    

def test_check_file_src_dir_input_pattern_init_file_returns_false():
    instance = get_dc_src_input()
    assert not instance.check_file(data_provider.get_change_data('/changelist_sort/__init__.py'))



"""The Changelist Map.
"""
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.list_key import ListKey


class ChangelistMap:
    """
    The Map containing all of the Changelists.
    """

    def __init__(self):
        self.mapping = {}
        self.changelist_ids = set()
    
    def insert(self, changelist: ChangelistData) -> bool:
        """
        Insert a Changelist into the Map.
        - Uses the Changelist Simple name as a key.

        Parameters:
        - changelist (ChangelistData): The Changelist Data to insert into the Map.

        Returns:
        bool - True when Changelist Name and Id are not already in the Map.
        """
        if changelist.id in self.changelist_ids or\
            changelist.list_key.key in self.mapping:
            return False
        self.changelist_ids.add(changelist.id)
        self.mapping[changelist.list_key.key] = changelist
        return True

    def search(self, key: str) -> ChangelistData | None:
        """
        Search the Map dict for the Changelist with the given simple name.
            Expects the Changelist Simple Name to match the key.
        """
        return self.mapping.get(key)

    def contains_id(self, id: str) -> bool:
        """
        Determine whether the Map contains the given id.
        """
        return id in self.changelist_ids

    def get_lists(self) -> list[ChangelistData]:
        """
        Obtain all Changelists in the Map as a List.
        """
        return list(self.mapping.values())

    def _generate_new_id(self) -> str:
        """
        Create a new Changelist Id that does not appear in this map.
        """
        from random import choices
        chars = list(_hex_char_generator())
        def generate_id():
            return '-'.join(''.join(choices(chars, k=x)) for x in (8, 4, 4, 4, 12))
        test_id = generate_id()
        while self.contains_id(test_id):
            test_id = generate_id()
        return test_id

    def create_changelist(self, cl_name: ListKey | str) -> ChangelistData:
        """
        Create a new empty Changelist with a new Id, and insert it into the Map.

        Parameters:
        - name (str): The Name of the new Changelist.

        Returns:
        ChangelistData - The Changelist that was recently created and added to the Map.
        """
        if isinstance(cl_name, ListKey):
            cl_name = cl_name.changelist_name 
        elif isinstance(cl_name, str):
            pass
        else:
            raise TypeError
        new_cl = ChangelistData(
            id=self._generate_new_id(),
            name=cl_name,
        )
        if self.insert(new_cl):
            return new_cl
        if (existing_cl := self.search(new_cl.list_key.key)) is not None:
            return existing_cl
        exit(f"Failed to create new Changelist(name={new_cl.list_key.changelist_name})")


def _hex_char_generator():
    """
    Generator yielding all possible hexadecimal characters.
    """
    for i in range(16):
        if i < 10:
            yield str(i) # Yield lowercase digits (0-9)
        else:
            yield chr(ord('a') + i - 10) # Yield uppercase letters (A-F)



"""Testing XML Reader Methods.
"""
from xml.etree.ElementTree import fromstring

from test.data_provider import get_multi_changelist_xml
from changelist_sort.workspace.xml_reader import filter_by_tag
from test.data_provider import get_no_changelist_xml, get_simple_changelist_xml


def test_filter_by_tag_no_changelist_returns_empty():
    root_elem = fromstring(get_no_changelist_xml())
    # There are no List Elements
    assert len(list(filter_by_tag(root_elem, 'list'))) == 0
    # There are no Change Elements
    assert len(list(filter_by_tag(root_elem, 'change'))) == 0
               

def test_filter_by_tag_simple_returns_1_list_1_change():
    root_elem = fromstring(get_simple_changelist_xml())
    # There are no List Elements
    assert len(list(filter_by_tag(root_elem, 'list'))) == 1
    # There are no Change Elements
    assert len(list(filter_by_tag(root_elem, 'change'))) == 1
               

def test_filter_by_tag_multi_returns_2_list_3_changes():
    root_elem = fromstring(get_multi_changelist_xml())
    # There are no List Elements
    assert len(list(filter_by_tag(root_elem, 'list'))) == 2
    # There are no Change Elements
    assert len(list(filter_by_tag(root_elem, 'change'))) == 3



""" Developer Files Definitions.
    Add New Types of Pattern Implementations Here.
"""
from typing import Callable

from changelist_sort.change_data import ChangeData


class DeveloperFilePattern:
    """
    A File Pattern to be controlled by the Developer.
    
    Properties:
    - module_type (ModuleType | None): Type of Module that this Pattern applies to.
    - _check_file (Callable[[ChangeData], bool]): Determines if the Change matches the pattern.
    """

    def __init__(self,
        inverse: bool = False,
        *args, **kwargs
    ):
        """
        Positional Arguments:
        - module_type (ModuleType | None): The Type of Module that this Pattern applies to. None applies to all.
        
        Keyword Arguments:
        - file_ext (str): The File Extension to Match.
        - first_dir (str): The First Directory in the File Path to Match.
        - filename_prefix (str): Match the Filename Prefix, ignoring first slash character.
        - filename_suffix (str): Match the Filename Suffix, ignoring file extension.
        - path_start (str): Match the start of the File Path, ignoring first slash character.
        - path_end (str): Match the end of the File Path Parent Directory, ignoring the end slash character.
        """
        self.inverse = inverse
        if 'file_ext' in kwargs:
            self._check_file = _match_file_ext(kwargs['file_ext'])
        elif 'first_dir' in kwargs:
            self._check_file = _match_first_dir(kwargs['first_dir'])
        elif 'filename_prefix' in kwargs:
            self._check_file = _match_filename_prefix(kwargs['filename_prefix'])
        elif 'filename_suffix' in kwargs:
            self._check_file = _match_filename_suffix(kwargs['filename_suffix'])
        elif 'path_start' in kwargs:
            self._check_file = _match_path_start(kwargs['path_start'])
        elif 'path_end' in kwargs:
            self._check_file = _match_path_end(kwargs['path_end'])
        # TODO: Add More Patterns To Constructor Keyword Arguments Here!
        # elif '' in kwargs:
        #   self._check_file = _match_your_new_pattern(kwargs[''])
        else:
            raise ValueError("One of the pattern matching keyword arguments must be provided.")

    def check_file(self, file: ChangeData) -> bool:
        """
        Determine if this File ChangeData matches the DeveloperFile Pattern.
        """
        return self.inverse.__xor__(self._check_file(file))


def _match_file_ext(
    file_ext: str,
) -> Callable[[ChangeData], bool]:
    """
    Match the given File Extension.
    """
    return lambda change_data: change_data.file_ext == file_ext


def _match_first_dir(
    first_dir: str | None,
) -> Callable[[ChangeData], bool]:
    """
    Match the given First Directory in the File Path.
    """
    return lambda change_data: change_data.first_dir == first_dir


def _match_filename_prefix(
    prefix: str,
) -> Callable[[ChangeData], bool]:
    """
    Match the Filename Prefix, ignoring first slash character.
    - Applies an internal function to overcome lambda limitations
    """
    return lambda change_data: change_data.file_basename.startswith(prefix)


def _match_filename_suffix(
    suffix: str,
) -> Callable[[ChangeData], bool]:
    """
    Match the Filename Suffix, ignoring file extension.
    - Applies an internal function to overcome lambda limitations
    """
    def get_filename(change_data: ChangeData) -> str:
        # Quickly return empty values
        if (filename := change_data.file_basename) == '':
            return ''
        if change_data.file_ext is not None:
            # Remove File Ext from basename
            filename = filename.removesuffix(change_data.file_ext)[:-1] # Remove file ext dot
        return filename
    # End Inner Function Definition
    return lambda change_data: get_filename(change_data).endswith(suffix)


def _match_path_start(
    path_start: str,
) -> Callable[[ChangeData], bool]:
    if not path_start.startswith('/'):
        return lambda change_data: change_data.sort_path[1:].startswith(path_start)
    else:
        return lambda change_data: change_data.sort_path.startswith(path_start)


def _match_path_end(
    path_end: str,
) -> Callable[[ChangeData], bool]:
    return lambda change_data: change_data.sort_path[:-len(change_data.file_basename)-1].endswith(path_end)


# TODO: Create Your Pattern Callables!
# Refer to ChangeData class for available pattern matching data
#
#def _match_your_new_pattern(
#    pattern_input: str,
#) -> Callable[[ChangeData], bool]:
#    return lambda change_data: False



""" Testing Main Package Init Module Methods.
"""
from pathlib import Path
from xml.etree.ElementTree import ElementTree

import pytest

from test import data_provider
from changelist_sort import sort_changelists
from changelist_sort.input.input_data import InputData
from changelist_sort.sorting.sort_mode import SortMode


def save_write(
    self: ElementTree, file_or_filename, encoding='utf-8', xml_declaration=True,
    method='xml'
):
    global TAG
    global VERSION
    elem = self.getroot()
    TAG = elem.tag
    VERSION = elem.attrib['version']


def test_sort_changelists_simple_module_sort_returns_xml():
    test_input = InputData(
        workspace_path=Path('file'),
        workspace_xml=data_provider.get_simple_changelist_xml(),
        sort_mode=SortMode.MODULE,
    )
    with (pytest.MonkeyPatch().context() as c):
        c.setattr(ElementTree, 'write', save_write)
        sort_changelists(test_input)
    assert TAG == 'project'
    assert VERSION == '4'



""" Testing List Sort Methods.
    The method split_changelist requires a callable arg, which applies the module sort for now.
"""
from test import data_provider

from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.sorting import module_sort
from changelist_sort.sorting.list_sort import split_changelist


def test_split_changelist_empty_returns_empty():
    test_input = ChangelistData(
        id='1234',
        name='Empty',
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 0
    assert len(result_list) == 0


def test_split_changelist_module_sorted_returns_sorted():
    test_input = ChangelistData(
        id='1234',
        name='Module',
        changes=[
            data_provider.get_module_src_change_data(),
            data_provider.get_module_test_change_data(),
        ],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 2
    assert len(result_list) == 0


def test_split_changelist_build_updates_sorted_returns_sorted():
    test_input = ChangelistData(
        id='1234',
        name='Build Updates',
        changes=[
            ChangeData(
                after_dir=False,
                after_path='/app/build.gradle',
            ),
        ],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 1
    assert len(result_list) == 0


def test_split_changelist_unsorted_returns_split():
    test_input = ChangelistData(
        id='1234',
        name='App',
        changes=[
            ChangeData(
                after_dir=False,
                after_path='/module/src/main/java/module/Main.java',
            ),
        ],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 0
    assert len(result_list) == 1


def test_split_changelist_half_sorted_returns_split():
    test_cd0 = ChangeData(
        after_dir=False,
        after_path='/module/src/main/java/module/Main.java',
    )
    test_cd1 = ChangeData(
        after_dir=False,
        after_path='/app/src/main/java/app/Main.java',
    )
    test_input = ChangelistData(
        id='1234',
        name='App',
        changes=[test_cd0, test_cd1],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 1
    assert test_input.changes[0] == test_cd1
    assert len(result_list) == 1
    assert result_list[0] == test_cd0


def test_split_changelist_root_unsorted_returns_unsorted():
    test_cd0 = ChangeData(
        after_dir=False,
        after_path='/app/Main.java',
    )
    test_input = ChangelistData(
        id='1234',
        name='Root',
        changes=[test_cd0],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 0
    assert len(result_list) == 1
    assert test_cd0 == result_list[0]


def test_split_changelist_project_root_sorted_returns_sorted():
    test_cd0 = ChangeData(
        after_dir=False,
        after_path='/Main.java',
    )
    test_input = ChangelistData(
        id='1234',
        name='Project Root',
        changes=[test_cd0],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 1
    assert test_input.changes[0] == test_cd0
    assert len(result_list) == 0


def test_split_changelist_root_sorted_returns_sorted():
    test_cd0 = ChangeData(
        after_dir=False,
        after_path='/Main.java',
    )
    test_input = ChangelistData(
        id='1234',
        name='Root',
        changes=[test_cd0],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 1
    assert test_input.changes[0] == test_cd0
    assert len(result_list) == 0


def test_split_changelist_build_updates_unsorted_returns_sorted():
    test_input = ChangelistData(
        id='1234',
        name='Build Updates',
        changes=[
            data_provider.get_module_src_change_data(),
            data_provider.get_module_test_change_data(),
            data_provider.get_github_workflows_change_data(),
            data_provider.get_github_dependabot_change_data(),
        ],
    )
    result_list = split_changelist(test_input, module_sort.is_sorted_by_module)
    assert len(test_input.changes) == 0
    assert len(result_list) == 4



"""Data describing a File Change.
"""
from os.path import basename
from dataclasses import dataclass, field


@dataclass(frozen=True)
class ChangeData:
    """The Change Information that is associated with a single file.

    Properties:
    - before_path (str | None): The initial path of the file.
    - before_dir (bool | None): Whether the initial file is a directory.
    - after_path (str | None): The final path of the file.
    - after_dir (bool | None): Whether the final path is a directory.
    """
    before_path: str | None = None
    before_dir: bool | None = None
    after_path: str | None = None
    after_dir: bool | None = None

    sort_path: str | None = field(init=False)
    first_dir: str | None = field(init=False, default=None)
    file_basename: str | None = field(init=False, default=None)    
    file_ext: str | None = field(init=False, default=None)

    def __post_init__(self):
        object.__setattr__(self, 'sort_path', self._get_sort_path())
        if self.sort_path is not None:
            object.__setattr__(self, 'first_dir', self._get_first_dir())
            object.__setattr__(self, 'file_basename', basename(self.sort_path))
            object.__setattr__(self, 'file_ext', self._get_file_ext())

    def _get_sort_path(self) -> str | None:
        """
        Determine the Path to use for sorting.
        """
        if self.before_path is not None:
            if self.after_path is not None:
                return self.after_path # prefer AfterPath
            else:
                return self.before_path
        elif self.after_path is not None:
            return self.after_path
        return None
    
    def _get_first_dir(self) -> str | None:
        """
        Obtain the First Directory in the file sort path.
        """
        try:
            if '/' == self.sort_path[0]:
                start_idx = 1
            else:
                start_idx = 0
            end_idx = self.sort_path.index('/', start_idx)
            return self.sort_path[start_idx:end_idx]
        except ValueError:
            return None

    def _get_file_ext(self) -> str | None:
        """
        Obtain the File Extension, or None.
        """
        try:
            return self.file_basename[self.file_basename.index('.', 1) + 1:]
        except ValueError:
            return None



""" Methods of Sorting Change Data.
    This Module aims to make Change Data easy to sort.
"""
from changelist_sort.change_data import ChangeData
from changelist_sort.sorting.module_type import ModuleType

_GRADLE_FILE_SUFFIXES = ('gradle', 'gradle.kts', 'properties')


def get_module_name(data: ChangeData) -> str | None:
    """
    Obtain the name of the module from the File path.
    """
    module_type = get_module_type(data)
    if module_type is None:
        return None
    if module_type == ModuleType.ROOT:
        return 'root'
    if module_type == ModuleType.GRADLE:
        return 'gradle'
    return data.first_dir.lower().lstrip('.')


def get_module_type(file: ChangeData) -> ModuleType | None:
    """
    Determine the ModuleType of the File.
    """
    if file.first_dir is None:
        if file.sort_path is None:
            return None
        if file.file_ext in _GRADLE_FILE_SUFFIXES:
            return ModuleType.GRADLE
        return ModuleType.ROOT
    if file.first_dir == 'gradle':
        return ModuleType.GRADLE
    if file.first_dir.startswith('.'):
        return ModuleType.HIDDEN
    if file.file_ext in _GRADLE_FILE_SUFFIXES:
        return ModuleType.GRADLE
    return ModuleType.MODULE



"""Setup Package Configuration
"""
from setuptools import setup, find_packages


setup(
    name="changelist-sort",
    version="0.2",
	author='DK96-OS',
	description='CLI Tool for Sorting Android Studio Changelists',
    long_description=open('README.md').read(),
    url="https://github.com/DK96-OS/changelist-sort/",
	project_urls={
        "Issues": "https://github.com/DK96-OS/changelist-sort/issues",
        "Source Code": "https://github.com/DK96-OS/changelist-sort/"
	},
	license='GPLv3',
    packages=find_packages(exclude=['test']),
    entry_points={
        'console_scripts': [
            'changelist-sort=changelist_sort.__main__:main',
            'changelist_sort=changelist_sort.__main__:main',
            'cl-sort=changelist_sort.__main__:main',
            'cl_sort=changelist_sort.__main__:main',
        ],
    },
    python_requires='>=3.10',
    classifiers=[
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: 3.12',
        'Programming Language :: Python :: 3.13',
    ],
)



""" Testing the Developer SortMode.
"""
from test import data_provider

from changelist_sort.sorting import sort
from changelist_sort.sorting.sort_mode import SortMode


def test_sort_dev_empty_returns_empty():
    test_input = []
    result = sort(test_input, SortMode.DEVELOPER)
    assert len(result) == 0


def test_sort_dev_build_updates_cl_sorted_returns_sorted():
    test_input = [data_provider.get_build_updates_changelist()]
    result = sort(test_input, SortMode.DEVELOPER)
    assert len(result) == 1
    assert len(result[0].changes) == 1



""" Testing Workspace Package Init Module Methods.
"""

from changelist_sort.input.input_data import InputData
from changelist_sort.workspace import get_workspace_tree


def test_get_workspace_tree_empty_str_raises_exit():
    test_input = InputData(
        workspace_xml='',
        workspace_path=None,
    )
    try:
        get_workspace_tree(test_input)
        assert False
    except SystemExit:
        assert True



""" Testing Developer Sort.
    This test module is for testing your developer file patterns and changelists.
    You should test the following two methods:
    - sort_file_by_developer
    - is_sorted_by_developer
    If you need to know which module type a file is, use helper method get_file_patterns
"""
from changelist_sort.change_data import ChangeData
from changelist_sort.sorting import developer_sort
from changelist_sort.sorting.module_type import ModuleType
from changelist_sort.workspace.developer_changelist import DeveloperChangelist
from test import data_provider

from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.list_key import ListKey
from changelist_sort.sorting import file_sort
from changelist_sort.sorting.developer_sort import sort_file_by_developer, is_sorted_by_developer


def get_file_patterns(file: ChangeData) -> tuple[DeveloperChangelist]:
    """
    Obtain the Developer Changelist Patterns that will be matched against this File Change.
    """
    module_type = file_sort.get_module_type(file)
    return developer_sort._filter_patterns_by_module(module_type)


def test_filter_patterns_by_module_root_returns_tuple():
    result = developer_sort._filter_patterns_by_module(ModuleType.ROOT)
    assert len(result) == 5
    # print(', '.join(x.list_key.changelist_name for x in result))


def test_filter_patterns_by_module_gradle_returns_tuple():
    result = developer_sort._filter_patterns_by_module(ModuleType.GRADLE)
    assert len(result) == 3
    # print(', '.join(x.list_key.changelist_name for x in result))


def test_filter_patterns_by_module_module_returns_tuple():
    result = developer_sort._filter_patterns_by_module(ModuleType.MODULE)
    assert len(result) == 9
    # print(', '.join(x.list_key.changelist_name for x in result))


def test_filter_patterns_by_module_hidden_returns_tuple():
    result = developer_sort._filter_patterns_by_module(ModuleType.HIDDEN)
    assert len(result) == 1
    # print(result[0].list_key.changelist_name)


def test_sort_file_by_developer_github_cl_exists_returns_true():
    cl_map = ChangelistMap()
    github_cl = data_provider.get_github_changelist()
    assert cl_map.insert(github_cl)
    test_file = data_provider.get_github_dependabot_change_data()
    assert sort_file_by_developer(cl_map, test_file)
    # Expect Same Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert new_cl.id == github_cl.id
    assert new_cl.name == github_cl.name
    assert new_cl.list_key == github_cl.list_key


def test_sort_file_by_developer_module_cl_creation_src_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_module_src_change_data()
    # Ensure the File is the right Module Type
    assert file_sort.get_module_type(test_file) == ModuleType.MODULE
    patterns_to_match = get_file_patterns(test_file)
    # The Number of Patterns with the given ModuleType
    assert len(patterns_to_match) == 9
    #
    assert sort_file_by_developer(cl_map, test_file)
    # Expect New Changelist
    result = cl_map.get_lists()
    new_cl = result[0]
    # The CL Key is the Module Name
    assert new_cl.list_key.key == file_sort.get_module_name(test_file)
    # Search for CL
    assert cl_map.search(new_cl.list_key.key) is not None


def test_sort_file_by_developer_gradle_module_app_build_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_app_gradle_build_change_data()
    #
    assert file_sort.get_module_type(test_file) == ModuleType.GRADLE
    assert sort_file_by_developer(cl_map, test_file)
    # Expect New Changelist
    result = cl_map.get_lists()
    new_cl = result[0]
    # The CL Key is the Module Name
    assert new_cl.list_key == developer_sort._BUILD_UPDATES_KEY
    # Search for CL
    assert cl_map.search(new_cl.list_key.key) is not None


def test_sort_file_by_developer_existing_gradle_module_app_build_file_returns_true():
    cl_map = ChangelistMap()
    cl_map.create_changelist(developer_sort._BUILD_UPDATES_KEY)
    test_file = data_provider.get_app_gradle_build_change_data()
    #
    assert file_sort.get_module_type(test_file) == ModuleType.GRADLE
    assert sort_file_by_developer(cl_map, test_file)
    # Expect New Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    # The CL Key is the Module Name
    assert new_cl.list_key == developer_sort._BUILD_UPDATES_KEY
    # Search for CL
    assert cl_map.search(new_cl.list_key.key) is not None


def test_is_sorted_by_developer_module_cl_creation_simple_key_returns_true():
    list_key = ListKey('module', 'Module Source Files')
    test_file = data_provider.get_module_src_change_data()
    # Ensure the File is the right Module Type
    assert file_sort.get_module_type(test_file) == ModuleType.MODULE
    #
    assert is_sorted_by_developer(list_key, test_file)


def test_is_sorted_by_developer_module_cl_creation_full_key_returns_true():
    list_key = ListKey('modulesourcefiles', 'Module Source Files')
    test_file = data_provider.get_module_src_change_data()
    # Ensure the File is the right Module Type
    assert file_sort.get_module_type(test_file) == ModuleType.MODULE
    #
    assert is_sorted_by_developer(list_key, test_file)


def test_is_sorted_by_developer_shell_script_build_file_returns_true():
    list_key = ListKey('shellscripts', 'Shell Scripts')
    test_file = data_provider.get_change_data('/shell_build.sh')
    # Ensure the File is the right Module Type
    assert file_sort.get_module_type(test_file) == ModuleType.ROOT
    #
    assert is_sorted_by_developer(list_key, test_file)


def test_is_sorted_by_developer_shellscripts_cl_non_shell_root_files_returns_false():
    list_key = ListKey('shellscripts', 'Shell Scripts')
    test_file = data_provider.get_change_data('/bat_build.bat')
    # Ensure the File is the right Module Type
    assert file_sort.get_module_type(test_file) == ModuleType.ROOT
    #
    assert not is_sorted_by_developer(list_key, test_file)



""" Testing Sorting Package Init Methods.
"""
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_data import ChangelistData
from changelist_sort.sorting import sort
from changelist_sort.sorting.sort_mode import SortMode
from test import data_provider


def test_sort_empty_returns_empty():
    test_input = []
    result = sort(test_input, SortMode.MODULE)
    assert result == test_input


def test_sort_empty_invalid_sort_mode_raises_exit():
    try:
        sort([], 'sort_mode.invalid')
        raised_exit = False
    except SystemExit:
        raised_exit = True
    assert raised_exit


def test_sort_invalid_sort_mode_raises_exit():
    test_input = [data_provider.get_app_changelist()]
    try:
        sort(test_input, 'sort_mode.invalid')
        raised_exit = False
    except SystemExit:
        raised_exit = True
    assert raised_exit


def test_sort_no_changes_returns_same():
    test_input = [
        ChangelistData(
            id='1234',
            name='Gradle',
            changes=[],
        )
    ]
    result = sort(test_input, SortMode.MODULE)
    assert result == test_input


def test_sort_gradle_cl_gradle_file_returns_unchanged():
    test_input = [
        ChangelistData(
            id='1234',
            name='Gradle',
            changes=[
                ChangeData(
                    after_path=data_provider.APP_GRADLE_PATH,
                    after_dir=False,
                ),
            ],
        )
    ]
    result = sort(test_input, SortMode.MODULE)
    assert result == test_input
    

def test_sort_gradle_cl_gradle_files_returns_unchanged():
    test_input = [
        ChangelistData(
            id='1234',
            name='Build Updates',
            changes=[
                ChangeData(
                    after_path=data_provider.APP_GRADLE_PATH,
                    after_dir=False,
                ),
                ChangeData(
                    after_path=data_provider.ROOT_GRADLE_PATH,
                    after_dir=False,
                ),
                ChangeData(
                    after_path=data_provider.GRADLE_PROPERTIES_PATH,
                    after_dir=False,
                ),
            ],
        )
    ]
    result = sort(test_input, SortMode.MODULE)
    assert result == test_input
    

def test_sort_app_cl_gradle_file_returns_new_gradle_cl():
    test_input = [
        ChangelistData(
            id='1234',
            name='App',
            changes=[
                data_provider.get_app_gradle_build_change_data(),
            ],
        )
    ]
    result = sort(test_input, SortMode.MODULE)
    assert result != test_input
    assert len(result) == 2
    assert result[0].name == 'App'
    assert result[1].name == 'Build Updates'
    # Check Contents of Changes List
    assert len(result[0].changes) == 0
    assert len(result[1].changes) == 1
    assert result[1].changes[0].after_path == data_provider.APP_GRADLE_PATH


def test_sort_app_cl_gradle_file_sourceset_mode_returns_new_gradle_cl():
    test_input = [
        ChangelistData(
            id='1234',
            name='App',
            changes=[
                data_provider.get_app_gradle_build_change_data(),
            ],
        )
    ]
    result = sort(test_input, SortMode.SOURCESET)
    assert result != test_input
    assert len(result) == 2
    assert result[0].name == 'App'
    assert result[1].name == 'Build Updates'
    # Check Contents of Changes List
    assert len(result[0].changes) == 0
    assert len(result[1].changes) == 1
    assert result[1].changes[0].after_path == data_provider.APP_GRADLE_PATH


def test_sort_module_cl_sourceset_mode_returns_new_sourceset_cl():
    test_input = [
        data_provider.get_module_changelist()
    ]
    result = sort(test_input, SortMode.SOURCESET)
    assert result != test_input
    assert len(result) == 3
    assert result[0].name == 'Module'
    assert result[1].name == 'Module Test'
    assert result[2].name == 'Module Main'
    # Check Files
    assert len(result[0].changes) == 0
    assert len(result[1].changes) == 1
    assert len(result[2].changes) == 1
    



""" Testing Changelist Data Methods
"""
from changelist_sort.changelist_data import ChangelistData


def test_list_key_empty_returns_empty():
    instance = ChangelistData(
        id='1234',
        name='',
        changes=[],
    )
    assert '' == instance.list_key.key


def test_list_key_space_returns_empty():
    instance = ChangelistData(
        id='1234',
        name=' ',
        changes=[],
    )
    assert '' == instance.list_key.key


def test_list_key_allcaps_returns_lower():
    instance = ChangelistData(
        id='1234',
        name='ALLCAPS',
        changes=[],
    )
    assert 'allcaps' == instance.list_key.key


def test_list_key_trailspace_returns_word():
    instance = ChangelistData(
        id='1234',
        name='Trailspace ',
        changes=[],
    )
    assert 'trailspace' == instance.list_key.key



"""String Validation Methods.
"""


def validate_name(argument) -> bool:
    """
    Determine whether an argument is a non-empty string.
        Does not count whitespace.
        Uses the strip method to remove empty space.

    Parameters:
    - argument (str) : The given argument.

    Returns:
    bool - True if the argument qualifies as valid.
    """
    if argument is None or not isinstance(argument, str):
        return False
    elif len(argument.strip()) < 1:
        return False
    return True




""" Testing Source Set Sorting Methods.
"""
from changelist_sort import list_key
from changelist_sort.changelist_data import ChangelistData
from test import data_provider
from changelist_sort.change_data import ChangeData
from changelist_sort.changelist_map import ChangelistMap
from changelist_sort.sorting.source_set_sort import _get_source_set_name, is_sorted_by_source_set, sort_by_source_set


_TEST_FIXTURES_CHANGE_PATH = '/module/src/testFixtures/java/module/MainTestFixture.java'
_ANDROID_TEST_CHANGE_PATH = '/module/src/androidTest/java/module/MainAndroidTest.java'


def test_sort_by_source_set_empty_map_empty_file_returns_false():
    cl_map = ChangelistMap()
    empty_file = ChangeData()
    assert not sort_by_source_set(cl_map, empty_file)
    

def test_sort_by_source_set_empty_map_module_src_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_module_src_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check the CL Name
    new_cl = cl_map.search('modulemain')
    assert new_cl is not None
    assert new_cl.name == 'Module Main'


def test_sort_by_source_set_empty_map_module_test_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_module_test_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check the CL Name
    new_cl = cl_map.search('moduletest')
    assert new_cl is not None
    assert new_cl.name == 'Module Test'


def test_sort_by_source_set_empty_map_module_debug_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_module_debug_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check the CL Name
    new_cl = cl_map.search('moduledebug')
    assert new_cl is not None
    assert new_cl.name == 'Module Debug'


def test_sort_by_source_set_empty_map_module_testfixtures_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_change_data(_TEST_FIXTURES_CHANGE_PATH)
    assert sort_by_source_set(cl_map, test_file)
    # Check the CL Name
    new_cl = cl_map.search('moduletestfixtures')
    assert new_cl is not None
    assert new_cl.name == 'Module Test Fixtures'


def test_sort_by_source_set_empty_map_module_androidtest_file_returns_true():
    cl_map = ChangelistMap()
    test_file = data_provider.get_change_data(_ANDROID_TEST_CHANGE_PATH)
    assert sort_by_source_set(cl_map, test_file)
    # Check the CL Name
    new_cl = cl_map.search('moduleandroidtest')
    assert new_cl is not None
    assert new_cl.name == 'Module Android Test'


def test_sort_by_source_set_app_gradle_file_returns_true_inserts_build_updates():
    cl_map = ChangelistMap()
    test_file = data_provider.get_app_gradle_build_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check CL Map for new Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert len(new_cl.changes) == 1
    assert new_cl.name == 'Build Updates'


def test_sort_by_source_set_github_workflows_returns_true_inserts_github():
    cl_map = ChangelistMap()
    test_file = data_provider.get_github_workflows_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check CL Map for new Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    new_cl = result[0]
    assert len(new_cl.changes) == 1
    assert new_cl.name == 'Github'


def test_sort_by_source_set_existing_module_cl_module_file_returns_true_inserts_module_main_cl():
    cl_map = ChangelistMap()
    # The Existing Module does not specify the Main source set
    assert cl_map.insert(data_provider.get_module_changelist([]))
    test_file = data_provider.get_module_src_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check CL Map for new ModuleMain Changelist and File
    result = cl_map.get_lists()
    assert len(result) == 2
    existing_module_cl = result[0]
    assert len(existing_module_cl.changes) == 0
    assert existing_module_cl.name == 'Module'
    new_main_cl = result[1]
    assert len(new_main_cl.changes) == 1
    assert new_main_cl.name == 'Module Main'


def test_sort_by_source_set_existing_module_cl_test_file_returns_true():
    cl_map = ChangelistMap()
    assert cl_map.insert(data_provider.get_module_changelist([]))
    test_file = data_provider.get_module_test_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check CL Map for Changelist and File
    result = cl_map.get_lists()
    assert len(result) == 2
    existing_module_cl = result[0]
    assert len(existing_module_cl.changes) == 0
    assert existing_module_cl.name == 'Module'
    new_test_cl = result[1]
    assert len(new_test_cl.changes) == 1
    assert new_test_cl.name == 'Module Test'


def test_sort_by_source_set_existing_moduletest_cl_moduletest_file_returns_true():
    cl_map = ChangelistMap()
    existing_test_cl = ChangelistData(
        id='2313411r325512',
        name='Module Test',
    )
    assert cl_map.insert(existing_test_cl)
    test_file = data_provider.get_module_test_change_data()
    assert sort_by_source_set(cl_map, test_file)
    # Check CL Map for File in Changelist
    result = cl_map.get_lists()
    assert len(result) == 1
    assert existing_test_cl == result[0]
    assert len(existing_test_cl.changes) == 1


def test_is_sorted_by_source_set_modulemain_cl_modulemain_file_returns_true():
    modulemain_key = list_key.compute_key('Module Main')
    test_file = data_provider.get_module_src_change_data()
    assert is_sorted_by_source_set(modulemain_key, test_file)
    

def test_is_sorted_by_source_set_module_cl_module_file_returns_false():
    module_key = list_key.compute_key('Module')
    test_file = data_provider.get_module_src_change_data()
    assert not is_sorted_by_source_set(module_key, test_file)
    

def test_is_sorted_by_source_set_moduletest_cl_moduletest_file_returns_true():
    moduletest_key = list_key.compute_key('Module Test')
    test_file = data_provider.get_module_test_change_data()
    assert is_sorted_by_source_set(moduletest_key, test_file)
    

def test_is_sorted_by_source_set_module_cl_modulemain_file_returns_false():
    module_key = list_key.compute_key('Module')
    test_file = data_provider.get_module_test_change_data()
    assert not is_sorted_by_source_set(module_key, test_file)


def test_is_sorted_by_source_set_appmain_cl_gradle_build_file_returns_false():
    app_key = list_key.compute_key('App Main')
    test_file = data_provider.get_app_gradle_build_change_data()
    assert not is_sorted_by_source_set(app_key, test_file)


def test_get_source_set_name_module_src_file_returns_src():
    test_file = data_provider.get_module_src_change_data()
    assert _get_source_set_name(test_file) == 'main'


def test_get_source_set_name_module_test_file_returns_test():
    test_file = data_provider.get_module_test_change_data()
    assert _get_source_set_name(test_file) == 'test'


def test_get_source_set_name_module_debug_file_returns_debug():
    test_file = data_provider.get_module_debug_change_data()
    assert _get_source_set_name(test_file) == 'debug'


def test_get_source_set_name_module_androidtest_file_returns_androidtest():
    test_file = data_provider.get_change_data(_ANDROID_TEST_CHANGE_PATH)
    assert _get_source_set_name(test_file) == 'androidTest'


def test_get_source_set_name_module_testfixtures_file_returns_testfixtures():
    test_file = data_provider.get_change_data(_TEST_FIXTURES_CHANGE_PATH)
    assert _get_source_set_name(test_file) == 'testFixtures'



""" Testing Developer File Pattern.
"""
from test import data_provider

from changelist_sort.workspace.developer_file_pattern import DeveloperFilePattern


def test_constructor_invalid_kwarg_raises_error():
    try:
        DeveloperFilePattern(invalid_kwarg='')
        raised_error = False
    except ValueError:
        raised_error = True
    assert raised_error


def test_check_file_file_ext_returns_true():
    instance = DeveloperFilePattern(
        inverse=False,
        file_ext='gradle'
    )
    assert instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )


def test_check_file_file_ext_inverse_returns_false():
    instance = DeveloperFilePattern(
        inverse=True,
        file_ext='gradle'
    )
    assert not instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )


def test_check_file_first_dir_app_app_gradle_returns_true():
    instance = DeveloperFilePattern(
        inverse=False,
        first_dir='app'
    )
    assert instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )
    

def test_check_file_first_dir_gradle_app_gradle_returns_false():
    instance = DeveloperFilePattern(
        inverse=False,
        first_dir='gradle'
    )
    assert not instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )
    

def test_check_file_first_dir_app_inverse_app_gradle_returns_false():
    instance = DeveloperFilePattern(
        inverse=True,
        first_dir='app'
    )
    assert not instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )
    

def test_check_file_first_dir_gradle_inverse_app_gradle_returns_true():
    instance = DeveloperFilePattern(
        inverse=True,
        first_dir='gradle'
    )
    assert instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )
    

def test_check_file_filename_prefix_build_app_gradle_returns_true():
    instance = DeveloperFilePattern(
        inverse=False,
        filename_prefix='build'
    )
    assert instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )


def test_check_file_filename_prefix_build_inverse_app_gradle_returns_false():
    instance = DeveloperFilePattern(
        inverse=True,
        filename_prefix='build'
    )
    assert not instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )
    

def test_check_file_filename_suffix_build_app_gradle_returns_true():
    instance = DeveloperFilePattern(
        inverse=False,
        filename_suffix='build'
    )
    assert instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )


def test_check_file_filename_suffix_build_inverse_app_gradle_returns_false():
    instance = DeveloperFilePattern(
        inverse=True,
        filename_suffix='build'
    )
    assert not instance.check_file(
        data_provider.get_app_gradle_build_change_data()
    )


def test_check_file_filename_suffix_file_basename_is_none_returns_false():
    instance = DeveloperFilePattern(
        filename_suffix='build'
    )
    cd = data_provider.get_change_data('/just/a/directory/')
    assert cd.file_basename == ''
    assert not instance.check_file(cd)


def test_check_file_path_start_github_github_workflows_returns_true():
    instance = DeveloperFilePattern(
        path_start='.github/'
    )
    assert instance.check_file(data_provider.get_github_workflows_change_data())


def test_check_file_path_start_github_github_dependabot_returns_true():
    instance = DeveloperFilePattern(
        path_start='.github/'
    )
    assert instance.check_file(data_provider.get_github_dependabot_change_data())


def test_check_file_path_start_slash_char_github_github_dependabot_returns_true():
    instance = DeveloperFilePattern(
        path_start='/.github/'
    )
    assert instance.check_file(data_provider.get_github_dependabot_change_data())


def test_check_file_path_end_github_workflows_returns_true():
    instance = DeveloperFilePattern(
        path_end='workflows'
    )
    assert instance.check_file(data_provider.get_github_workflows_change_data())


def test_check_file_path_end_workflows_github_workflows_returns_true():
    instance = DeveloperFilePattern(
        path_end='/workflows'
    )
    assert instance.check_file(data_provider.get_github_workflows_change_data())


def test_check_file_path_end_workflows_github_dependabot_returns_false():
    instance = DeveloperFilePattern(
        path_end='/workflows'
    )
    assert not instance.check_file(data_provider.get_github_dependabot_change_data())
